#ifndef _REPLAY_H

using namespace std;

#include <vamos/geometry/Three_Vector.h>
#include <vamos/geometry/Three_Matrix.h>
#include <vamos/body/Car.h>

#include "serialization.h"

#define DBG_REPLAY false

class REPLAYCARSTATE
{
private:
friend class VAMOSWORLD;
friend class SERIALIZER;

	int frameno;
	
	Vamos_Geometry::Three_Vector chassispos;
	Vamos_Geometry::Three_Matrix chassisorientation;
	Vamos_Geometry::Three_Vector chassisvel;
	Vamos_Geometry::Three_Vector chassisangvel;
	double suspdisp[4];
	double suspcompvel[4];
	Vamos_Geometry::Three_Vector whlangvel[4];
	int gear;
	double enginespeed;
	double clutchspeed;
	double enginedrag;
	double tirespeed[4];

	int segment;
	
public:
	void Set(const int newframeno, const int newseg, Vamos_Body::Car & car);
	int GetFrame() const {return frameno;}
	
	void Serialize(SERIALIZER * s)
	{
		_SERIALIZE_(s,frameno);
		_SERIALIZE_(s,chassispos);
		_SERIALIZE_(s,chassisorientation);
		_SERIALIZE_(s,chassisvel);
		_SERIALIZE_(s,chassisangvel);
		for (int i = 0; i < 4; i++)
		{
			_SERIALIZE_(s,suspdisp[i]);
			_SERIALIZE_(s,suspcompvel[i]);
			_SERIALIZE_(s,whlangvel[i]);
			_SERIALIZE_(s,tirespeed[i]);
		}
		_SERIALIZE_(s,gear);
		_SERIALIZE_(s,enginespeed);
		_SERIALIZE_(s,clutchspeed);
		_SERIALIZE_(s,enginedrag);
		_SERIALIZE_(s,segment);
	}
};

class REPLAYINPUT
{
	private:
		friend class SERIALIZER;
		float value;
		int id;
		int helduntil;
	public:
		REPLAYINPUT() : value(0),id(0),helduntil(0) {}
		REPLAYINPUT(float newval, int newid, int newheld) : value(newval), id(newid), helduntil(newheld) {}
		REPLAYINPUT(const REPLAYINPUT & other)
		{
			//just use the default assignment operator, which will do a shallow copy
			*this = other;
		}
		
		void Serialize(SERIALIZER * s)
		{
			_SERIALIZE_(s,value);
			_SERIALIZE_(s,id);
			_SERIALIZE_(s,helduntil);
		}
		
		void HoldUntil(int newhu) {helduntil = newhu;}
		int GetHeldUntil() {return helduntil;}
		int GetID() {return id;}
		float GetValue() {return value;}
};

class REPLAYINPUTFRAME
{
	private:
		friend class SERIALIZER;
		int frameno;
		list <REPLAYINPUT> inputlist;
		list <REPLAYINPUT>::iterator it;
		
	public:
		REPLAYINPUTFRAME() : frameno(0) {}
		REPLAYINPUTFRAME(const int newframeno) : frameno(newframeno) {}
		REPLAYINPUTFRAME(const REPLAYINPUTFRAME & other)
		{
			//just use the default assignment operator, which will do a shallow copy
			*this = other;
		}
		
		REPLAYINPUT & AddInput(const REPLAYINPUT newinput)
		{
			inputlist.push_back(newinput);
			return inputlist.back();
		}
		
		void Serialize(SERIALIZER * s)
		{
			_SERIALIZE_(s,frameno);
			_SERIALIZELIST_(s,inputlist);
		}
		
		int GetFrame() const {return frameno;}
		
		void ResetIterator() {it = inputlist.begin();}
		bool IteratorAtEnd() {return (it == inputlist.end());}
		void IncrementIterator() {it++;}
		REPLAYINPUT & GetIteratorInput() {return *it;}
};

class REPLAY
{
private:
friend class SERIALIZER;

	const string replay_version;
	float frametime;
	int totalframes;
	
	bool ghostcar;
	
	string replaytrack;
	string replaycar;
	int replaypaint;
	bool replayabs;
	bool replaytcs;
	
	string replayfilename;
	
	int frame;
	
	//these iterators are used during playback to keep track of the current position
	list <REPLAYINPUTFRAME>::iterator inputframelist_iterator;
	list <REPLAYCARSTATE>::iterator statelist_iterator;
	
	list <REPLAYINPUTFRAME> inputframelist; ///< chronological list of saved inputs
	list <REPLAYCARSTATE> statelist; ///< chronological list of saved car states
	map <int, string> function_id_map; ///< maps function ids and function strings
	map <int, pair <float, REPLAYINPUT*> > last_values; ///< maps function ids to their last/cached values
	list <REPLAYINPUT> playback_heldbuttons; ///< a list of last_values still waiting to be applied this frame (during playback)
	
	enum
	{
		REPLAY_INACTIVE,
		REPLAY_RECORDING,
  		REPLAY_PLAYING
	} mode;
	
	void WriteRecording(string replay_filename);
	void RecordingFrameEnd();
	void PlayingFrameStart();

public:
	REPLAY() : replay_version("VDRIFTREPLAYV08"),totalframes(0),ghostcar(false),frame(0),mode(REPLAY_INACTIVE) {}
	
	void Serialize(SERIALIZER * s)
	{
		_SERIALIZE_(s,totalframes);
		
		_SERIALIZE_(s,replaytrack);
		_SERIALIZE_(s,replaycar);
		_SERIALIZE_(s,replayabs);
		_SERIALIZE_(s,replaytcs);
		_SERIALIZE_(s,replaypaint);
		
		_SERIALIZEMAP_(s,function_id_map);
		_SERIALIZELIST_(s,statelist);
		_SERIALIZELIST_(s,inputframelist);
	}
	
	//adjust playback time
	void Jump(double newtime);
	
	//get playback or recording elapsed time
	float GetTime() {if (Recording() || Playing()) return frametime * frame; else return 0.0;}
	
	//get playback total time
	float GetTotalTime() {if (!Playing()) return 0; else return frametime * totalframes;}
	
	//this block of functions is used to add recording info
	void AddRecord(string newdofunction, float newval, int newstate);
	void AddCarState(int newsegment, Vamos_Body::Car & newstate);
	
	//used for both playback and recording
	void IncrementFrame();
	
	//this block of functions is used for playback
	const REPLAYCARSTATE GetState();
	bool StateAvailable() const;
	void GetOp(string & output_opname, float & output_opvalue);
	bool OpAvailable() const;
	void IncrementOp();
	
	//control playing
	bool StartPlaying(string replayname, float newframetime, bool ngc=false, bool collect_bench_data=false);
	void StopPlaying() {if (!Playing()) return; mode = REPLAY_INACTIVE;}
	bool Playing() const {return (mode == REPLAY_PLAYING);}
	
	//this block of functions is used for playback to determine settings
	bool GhostCar() {return ghostcar;}
	bool ReplayABS() {return replayabs;}
	bool ReplayTCS() {return replaytcs;}
	string ReplayCar() {return replaycar;}
	int ReplayPaint() {return replaypaint;}
	string ReplayTrack() {return replaytrack;}
	
	//control recording
	void StartRecording(string newreplayfilename, bool abs, bool tcs, int paint, string car, string track, float newframetime);
	void StopRecording();
	bool Recording() const {return (mode == REPLAY_RECORDING);}
};

#define _REPLAY_H
#endif /* _REPLAY_H */
