#include "replay.h"
#include "serialization.h"
#include "exception.h"

#include <sstream>

void REPLAY::WriteRecording(string replay_filename)
{
	ofstream replaystream(replay_filename.c_str());
	SERIALIZER_BIN_OUTPUT serialize_output(replaystream);
	serialize_output.SetPointerTracking(false);
	
	//write header
	string version(replay_version);
	serialize_output.Serialize("version", version);
	serialize_output.Serialize("frametime", frametime);
	
	Serialize(&serialize_output);
	
	
	
	/*ofstream debug_output((replay_filename+".xml").c_str());
	SERIALIZER_XML_OUTPUT xml(debug_output);
	xml.SetPointerTracking(false);
	xml.Serialize("version", version);
	xml.Serialize("frametime", frametime);
	Serialize(&xml);*/
}

void REPLAY::AddRecord(string newdofunction, float newval, int newstate)
{
	//if (newdofunction == "clutch") cout << newval << endl;
	
	//check to see if we have the function mapped to an id yet.  if not, do it
	int id = -1;
	for (map <int, string>::iterator i = function_id_map.begin(); i != function_id_map.end(); ++i)
	{
		if (i->second == newdofunction)
		{
			id = i->first;
		}
	}
	if (id == -1)
	{
		id = function_id_map.size();
		function_id_map[id] = newdofunction;
	}
	
	//check to see if we have a cached/last value for this function yet.  if not, create one
	pair <map <int, pair <float, REPLAYINPUT*> >::iterator, bool> lastval = last_values.insert(pair <int,pair <float, REPLAYINPUT *> > (id,pair <float, REPLAYINPUT *>(0,NULL)));
	
	//if the cached/last value is different, create a new entry and update the cached/last value
	if (lastval.second || lastval.first->second.first != newval)
	{
		//create a replayinputframe for this frame
		if (inputframelist.empty())
		{
			inputframelist.push_back(REPLAYINPUTFRAME(frame));
		}
		else
		{
			if (inputframelist.back().GetFrame() != frame)
				inputframelist.push_back(REPLAYINPUTFRAME(frame));
		}
	
		//by now, we are assured that inputframelist.back() exists and is correct for this frame
		
		//create a new input record and assign our held function tracking pointer
		lastval.first->second.second = &(inputframelist.back().AddInput(REPLAYINPUT(newval, id, frame)));
		
		//create a new last value entry
		lastval.first->second.first = newval;
		
	}
	else if (!lastval.second && lastval.first->second.first == newval)
	{
		//the value already exists in our cache map, and it's identical to what's already there
		if (!lastval.first->second.second)
			throw VDrift::EXCEPTION(__FILE__,__LINE__,"Corruption in replay recording memory");
		
		lastval.first->second.second->HoldUntil(frame);
	}
}

///clear out our held function map if the function didn't get activated this frame
void REPLAY::RecordingFrameEnd()
{
	list <int> to_erase;
	
	for (map <int, pair <float, REPLAYINPUT*> >::iterator i = last_values.begin(); i != last_values.end(); i++)
	{
		//clear out anything in the map that hasn't been held until this frame
		if (!i->second.second)
			throw VDrift::EXCEPTION(__FILE__,__LINE__,"Corruption in replay recording memory");
		
		if (i->second.second->GetHeldUntil() != frame)
			to_erase.push_back(i->first);
	}
	
	for (list <int>::iterator i = to_erase.begin(); i != to_erase.end(); i++)
	{
		last_values.erase(*i);
	}
}

void REPLAY::AddCarState(int newsegment, Vamos_Body::Car & newstate)
{
	//check time
	int state_interval = (int)(1.0 / frametime); //save a state every second
	if (frame % state_interval != 0)
		return;
	
	//check to make sure the most recent state isn't already in the list
	if (!statelist.empty())
	{
		if (statelist.back().GetFrame() == frame)
			return;
	}
	
	//add a new state to the list
	statelist.push_back(REPLAYCARSTATE());
	statelist.back().Set(frame, newsegment, newstate);
}

void REPLAYCARSTATE::Set(const int newframeno, const int newseg, Vamos_Body::Car & car)
{
	frameno = newframeno;
	segment = newseg;
	
	car.GetState(chassispos,
		chassisorientation,
		chassisvel,
		chassisangvel,
		suspdisp,
		suspcompvel,
		whlangvel,
		gear,
		enginespeed,
		clutchspeed,
		enginedrag,
		tirespeed);
}

void REPLAY::StartRecording(string newreplayfilename, bool abs, bool tcs, int paint, string car, string track, float newframetime)
{
	if (Playing())
		StopPlaying();
	
	frametime = newframetime;
	replayabs = abs;
	replaytcs = tcs;
	replaypaint = paint;
	replaycar = car;
	replaytrack = track;
	frame = 0;
	mode = REPLAY_RECORDING;
	function_id_map.clear();
	last_values.clear();
	replayfilename = newreplayfilename;
}

void REPLAY::StopRecording()
{
	if (!Recording())
		return;
	totalframes = frame;
	mode = REPLAY_INACTIVE;
	WriteRecording(replayfilename);
	totalframes = 0;
	ghostcar = false;
	frame = 0;
	
	//clear out replay data
	inputframelist.clear();
	statelist.clear();
	function_id_map.clear();
	last_values.clear();
	playback_heldbuttons.clear();
}

void REPLAY::IncrementFrame()
{
	if (Recording())
		RecordingFrameEnd();
	
	if (Playing() || Recording())
		frame++;
	
	if (frame > 2000000000)
	{
		StopPlaying();
		StopRecording();
	}
	
	if (Playing())
		PlayingFrameStart();
}

bool REPLAY::StartPlaying(string replayname, float newframetime, bool ngc, bool collect_bench_data)
{
	if (Recording())
		StopRecording();
	
	string replay_filename(replayname);
	//replay_filename = "test.vdr";
	ifstream replaystream(replay_filename.c_str());
	if (!replaystream)
		return false;
	
	/*if (replaystream.peek() != 15)
	{
		cerr << "Replay version string is incorrect" << endl;
		cout << replaystream.get() << "," << replaystream.get() << "," << replaystream.get() << endl;
		return false;
	}*/
	
	SERIALIZER_BIN_INPUT serialize_input(replaystream);
	serialize_input.SetPointerTracking(false);
	
	//read header
	string version(replay_version);
	serialize_input.Serialize("version", version);
	
	if (version != replay_version)
	{
		cerr << "Replay version of " << version << " doesn't match supported version " << replay_version << endl;
		return false;
	}
	
	serialize_input.Serialize("frametime", frametime);
	
	if (frametime != newframetime)
	{
		cerr << "Replay frametime of " << frametime << " doesn't match current frametime of " << newframetime << endl;
		return false;
	}
	
	Serialize(&serialize_input);
	
	last_values.clear();
	playback_heldbuttons.clear();
	frame = 0;
	mode = REPLAY_PLAYING;
	
	statelist_iterator = statelist.begin();
	inputframelist_iterator = inputframelist.begin();
	if (inputframelist_iterator != inputframelist.end())
		inputframelist_iterator->ResetIterator();
		
	return true;
}

void REPLAY::PlayingFrameStart()
{
	if (frame >= totalframes)
		StopPlaying();
	else
	{
		if (!playback_heldbuttons.empty()) //should be emptied by OpIncrement()
			throw VDrift::EXCEPTION(__FILE__,__LINE__,"Corruption in replay playback memory");
		
		list <int> to_erase;
		for (map <int, pair <float, REPLAYINPUT*> >::iterator i = last_values.begin(); i != last_values.end(); i++)
		{
			if (!i->second.second)
				throw VDrift::EXCEPTION(__FILE__,__LINE__,"Corruption in replay playback memory");
			
			//clear out old last_values that no longer apply
			if (i->second.second->GetHeldUntil() < frame)
			{
				to_erase.push_back(i->first);
			}
			else
			{
				//if it still applies, queue it up in playback_heldbuttons
				playback_heldbuttons.push_back(*(i->second.second));
			}
		}
		
		//do deletion of last_value items that no longer apply
		for (list <int>::iterator i = to_erase.begin(); i != to_erase.end(); i++)
		{
			last_values.erase(*i);
		}
	}
}

void REPLAY::Jump(double newtime)
{
	
}

///ensure that REPLAY::StateAvailable is true before calling this function
const REPLAYCARSTATE REPLAY::GetState()
{
	if (!StateAvailable())
	{
		//uh oh, someone called GetState without checking for StateAvailable()
		throw VDrift::EXCEPTION(__FILE__,__LINE__,"Corruption in replay playback memory");
		return REPLAYCARSTATE();
	}
	
	//StateAvailable already checks to ensure statelist_iterator is valid for this frame
	REPLAYCARSTATE & thisstate = *statelist_iterator;
	
	statelist_iterator++;
	
	return thisstate;
}

///returns true if the statelist_iterator is pointing to a valid state for this frame
bool REPLAY::StateAvailable() const
{
	if (!Playing() || statelist_iterator == statelist.end())
		return false;
	
	if (statelist_iterator->GetFrame() == frame || (statelist_iterator->GetFrame() == 0 && frame == 1))
		return true;
	else if (statelist_iterator->GetFrame() == 0)
		return false;
	else if (statelist_iterator->GetFrame() < frame)
	{
		//uh oh, something weird happened.. we're pointing to an old state
		cerr << totalframes << endl;
		stringstream msg;
		msg << "Corruption in replay playback memory: " << statelist_iterator->GetFrame() << " is less than " << frame;
		throw VDrift::EXCEPTION(__FILE__,__LINE__,msg.str());
	}
	else //the next state is in the future
		return false;
}

void REPLAY::GetOp(string & output_opname, float & output_opvalue)
{
	if (!OpAvailable())
	{
		//uh oh, someone called this function without checking for OpAvailable()
		throw VDrift::EXCEPTION(__FILE__,__LINE__,"Corruption in replay playback memory");
	}
	
	int opid(-1);
	float opval(0);
	
	//run through held buttons first
	if (!playback_heldbuttons.empty())
	{
		opid = playback_heldbuttons.back().GetID();
		opval = playback_heldbuttons.back().GetValue();
	}
	else //no more held buttons; process new buttons
	{
		if (inputframelist_iterator != inputframelist.end())
		{
			if (!inputframelist_iterator->IteratorAtEnd())
			{
				opid = inputframelist_iterator->GetIteratorInput().GetID();
				opval = inputframelist_iterator->GetIteratorInput().GetValue();
				
				//add new stuff to last_values
				last_values[opid] = pair <float, REPLAYINPUT *> (opval, &inputframelist_iterator->GetIteratorInput());
				
			}
			else
				throw VDrift::EXCEPTION(__FILE__,__LINE__,"Corruption in replay playback memory");
		}
		else
			throw VDrift::EXCEPTION(__FILE__,__LINE__,"Corruption in replay playback memory");
	}
	
	output_opvalue = opval;
	map <int, string>::iterator i = function_id_map.find(opid);
	if (i == function_id_map.end())
		throw VDrift::EXCEPTION(__FILE__,__LINE__,"Corruption in replay playback memory");
	output_opname = i->second;
}

bool REPLAY::OpAvailable() const
{
	if (!Playing())
		return false;
	
	if (!playback_heldbuttons.empty())
		return true;
	
	if (inputframelist_iterator == inputframelist.end())
		return false;
	
	if (inputframelist_iterator->GetFrame() == frame || (inputframelist_iterator->GetFrame() == 0 && frame == 1))
		return true;
	else if (inputframelist_iterator->GetFrame() < frame)
	{
		//uh oh, something weird happened.. we're pointing to an old replayinputframe
		stringstream msg;
		msg << "Corruption in replay playback memory: " << inputframelist_iterator->GetFrame() << " is less than " << frame;
		throw VDrift::EXCEPTION(__FILE__,__LINE__,msg.str());
	}
	else //the next replayinputframe is in the future
		return false;
}

void REPLAY::IncrementOp()
{
	//increment through the playback_heldbuttons list
	if (!playback_heldbuttons.empty())
	{
		playback_heldbuttons.pop_back();
	}
	else //the playback_heldbuttons list is empty
	{		
		//throw if we're out of inputframes (should have been checked by OpAvailable)
		if (inputframelist_iterator == inputframelist.end())
		{
			throw VDrift::EXCEPTION(__FILE__,__LINE__,"Corruption in replay playback memory");
			return;
		}
		
		if (!inputframelist_iterator->IteratorAtEnd())
		{
			//increment inputlist iterator
			inputframelist_iterator->IncrementIterator();
		}
		
		//if out of events this frame, go to the next
		if (inputframelist_iterator->IteratorAtEnd())
		{
			inputframelist_iterator++;
			if (inputframelist_iterator != inputframelist.end())
			{
				inputframelist_iterator->ResetIterator();
				
				//throw if the inputframelist has no inputs
				if (inputframelist_iterator->IteratorAtEnd())
				{
					throw VDrift::EXCEPTION(__FILE__,__LINE__,"Corruption in replay playback memory");
				}
			}
		}
	}
}
